from .ReactionArchtype import ReactionArchtype

michaelis_menten = ReactionArchtype(
    'Michaelis Menten',
    ('&S',), ('&E',),
    ('Km', 'Vmax'),
    'Vmax*&S/(Km + &S)',
    assume_parameters_values={'Km': 100, 'Vmax': 10},
    assume_reactant_values={'&S': 100},
    assume_product_values={'&E': 0})

michaelis_menten_fixed = ReactionArchtype(
    'Michaelis Menten',
    ('&S',), ('&E',),
    (),
    '100*&S/(1 + &S)',
    assume_reactant_values={'&S': 100},
    assume_product_values={'&E': 0})

mass_action_21 = ReactionArchtype(
    'Mass Action',
    ('&A', '&B'), ('&C',),
    ('ka', 'kd'),
    'ka*&A*&B',
    assume_parameters_values={'ka': 0.001, 'kd': 0.01},
    assume_reactant_values={'&A': 100, '&B': 100},
    assume_product_values={'&C': 0},
    reversible=True,
    reverse_rate_law='kd*&C')

simple_rate_law = ReactionArchtype(
    'Simple Rate Law',
    ('&A',), ('&B',),
    ('kf',),
    'kf*&A',
    assume_parameters_values={'kf': 0.01},
    assume_reactant_values={'&A': 100},
    assume_product_values={'&B': 0})

michaelis_menten_stim = ReactionArchtype(
    'Michaelis Menten',
    ('&S',), ('&E',),
    ('Km', 'Vmax'),
    'Vmax*&S*&I/(Km + &S)',
    extra_states=('&I',),
    assume_parameters_values={'Km': 100, 'Vmax': 10},
    assume_reactant_values={'&S': 100},
    assume_product_values={'&E': 0})

michaelis_menten_inh_allosteric = ReactionArchtype(
    'Michaelis Menten',
    ('&S',), ('&E',),
    ('Km', 'Vmax', 'Ki'),
    'Vmax*&S/(Km + &S)*(1+&I*Ki)',
    extra_states=('&I',),
    assume_parameters_values={'Km': 100, 'Vmax': 10, 'Ki': 0.1},
    assume_reactant_values={'&S': 100},
    assume_product_values={'&E': 0})

michaelis_menten_inh_competitive_1 = ReactionArchtype(
    'Michaelis Menten',
    ('&S',), ('&E',),
    ('Km', 'Vmax', 'Ki'),
    'Vmax*&S/(Km*(1+&I*Ki) + &S)',
    extra_states=('&I',),
    assume_parameters_values={'Km': 100, 'Vmax': 10, 'Ki': 0.1},
    assume_reactant_values={'&S': 100},
    assume_product_values={'&E': 0})

degredation = ReactionArchtype(
    'Degredation',
    ('&A',), (),
    ('Kdeg',),
    'Kdeg*&A',
    assume_parameters_values={'Kdeg': 0.01},
    assume_reactant_values={'&A': 100},
    assume_product_values={})

synthesis = ReactionArchtype(
    'Synthesis',
    (), ('&A',),
    ('Ksyn',),
    'Ksyn',
    assume_parameters_values={'Ksyn': 0.1},
    assume_reactant_values={},
    assume_product_values={'&A': 0})

def create_archtype_synthesis(allo_stimulators=0, allo_inhibitors=0):
    reactants = ()
    products = ('&R',)
    parameters = ('Ksyn',)
    rate_law = 'Ksyn'
    extra_states = ()
    assume_reactant_values = {}
    assume_product_values = {'&R': 0}
    assume_parameters_values = {'Ksyn': 0.1}
    total_extra_states = ()
    if allo_stimulators > 0:
        # add the stimulators to the equation
        stim_str = '*('
        for i in range(allo_stimulators):
            stim_str += f'&A{i}*Ks{i}+'
        
        rate_law += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'&A{i}' for i in range(allo_stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ks{i}' for i in range(allo_stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ks{i}': 1e-4 for i in range(allo_stimulators)})

    if allo_inhibitors > 0:
        # add the inhibitors to the equation
        inh_str = '*(1/(1+'
        for i in range(allo_inhibitors):
            inh_str += f'&I{i}*Ki{i}+'
        
        rate_law += inh_str[:-1] + '))'

        # fill extra states
        extra_states += tuple([f'&I{i}' for i in range(allo_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ki{i}' for i in range(allo_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ki{i}': 1e-4 for i in range(allo_inhibitors)})

    return ReactionArchtype(
        'Synthesis',
        reactants, products,
        parameters,
        rate_law,
        extra_states=extra_states,
        assume_reactant_values=assume_reactant_values,
        assume_product_values=assume_product_values,
        assume_parameters_values=assume_parameters_values)

def create_archtype_degredation(allo_stimulators=0, allo_inhibitors=0):
    """
    Creates a degredation reaction archtype with the given number of
    stimulators and inhibitors.
    """
    reactants = ('&R',)
    products = ()
    parameters = ('Kdeg',)
    rate_law = 'Kdeg*&R'
    assume_reactant_values = {'&R': 100}
    assume_product_values = {}
    extra_states = ()
    assume_parameters_values = {'Kdeg': 0.01}
    total_extra_states = ()

    if allo_stimulators > 0:
        # add the stimulators to the equation
        stim_str = '*('
        for i in range(allo_stimulators):
            stim_str += f'&A{i}*Ks{i}+'
        
        rate_law += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'&A{i}' for i in range(allo_stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ks{i}' for i in range(allo_stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ks{i}': 1e-4 for i in range(allo_stimulators)})

    if allo_inhibitors > 0:
        # add the inhibitors to the equation
        inh_str = '*(1/(1+'
        for i in range(allo_inhibitors):
            inh_str += f'&I{i}*Ki{i}+'
        
        rate_law += inh_str[:-1] + '))'

        # fill extra states
        extra_states += tuple([f'&I{i}' for i in range(allo_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ki{i}' for i in range(allo_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ki{i}': 1e-4 for i in range(allo_inhibitors)})


    return ReactionArchtype(
        'Degredation General',
        reactants, products,
        parameters,
        rate_law,
        extra_states=extra_states,
        assume_parameters_values=assume_parameters_values,
        assume_reactant_values=assume_reactant_values,
        assume_product_values=assume_product_values)

def create_archtype_mass_action(
                    reactant_count=1, 
                    product_count=1, 
                    allo_stimulators=0, 
                    additive_stimulators=0, 
                    allo_inhibitors=0, 
                    comp_inhibitors=0,
                    rev_allo_stimulators=0,
                    rev_additive_stimulators=0,
                    rev_allo_inhibitors=0,
                    rev_comp_inhibitors=0,):
    reactants = tuple(f'&R{i}' for i in range(reactant_count))
    products = tuple(f'&P{i}' for i in range(product_count))
    parameters = ('Ka', 'Kd')
    assume_parameters_values={'Ka': 0.001, 'Kd': 0.01}
    forw_rate_law = 'Ka'
    for i in range(reactant_count):
        forw_rate_law += f'*&R{i}'
    rev_rate_law = 'Kd'
    for i in range(product_count):
        rev_rate_law += f'*&P{i}'

    total_extra_states = ()

    if allo_stimulators > 0:
        # add the stimulators to the equation
        stim_str = '*('
        for i in range(allo_stimulators):
            stim_str += f'&A{i}*Ks{i}+'
        
        forw_rate_law += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'&A{i}' for i in range(allo_stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ks{i}' for i in range(allo_stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ks{i}': 1e-4 for i in range(allo_stimulators)})

    if additive_stimulators > 0:
        # weak stimulators represent that stimulant is not required for the reaction to occur
        stim_weak_str = '(Ka+'
        for i in range(additive_stimulators):
            stim_weak_str += f'&W{i}*Kw{i}+'
        
        forw_rate_law = stim_weak_str[:-1] + ')' + forw_rate_law[2:]

        # fill extra states
        extra_states = tuple([f'&W{i}' for i in range(additive_stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kw{i}' for i in range(additive_stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kw{i}': 1e-4 for i in range(additive_stimulators)})

    if allo_inhibitors > 0:
        # add the inhibitors to the equation, exact same thing as allo_stimulators but 
        # applied to the reverse rate law, with different parameter names 

        # add the stimulators to the equation
        altered_str = '*(1/(1+'
        for i in range(allo_inhibitors):
            altered_str += f'&I{i}*Ki{i}+'

        forw_rate_law += altered_str[:-1] + '))'

        # fill extra states
        extra_states = tuple([f'&I{i}' for i in range(allo_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ki{i}' for i in range(allo_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ki{i}': 1e-4 for i in range(allo_inhibitors)})

    if comp_inhibitors > 0:
        # add the inhibitors to the equation, exact same thing as addi_stimulators but
        # applied to the reverse rate law, with different parameter names

        assert additive_stimulators == 0, 'Currently this function does not support having both additive and competitive inhibitors'

        # add the stimulators to the equation
        stim_str = '(Ka/(1+'
        for i in range(comp_inhibitors):
            stim_str += f'&C{i}*Kc{i}+'

        forw_rate_law = stim_str[:-1] + '))' + forw_rate_law[2:]

        # fill extra states
        extra_states = tuple([f'&C{i}' for i in range(comp_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kc{i}' for i in range(comp_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kc{i}': 1e-4 for i in range(comp_inhibitors)})

    # reverse rate law MODIFICATIONS

    if rev_allo_stimulators > 0:
        # add the stimulators to the equation
        stim_str = '*('
        for i in range(rev_allo_stimulators):
            stim_str += f'?A{i}*Ksr{i}+'
        
        rev_rate_law += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'?A{i}' for i in range(rev_allo_stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ksr{i}' for i in range(rev_allo_stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'rKs{i}': 1e-4 for i in range(rev_allo_stimulators)})

    if rev_additive_stimulators > 0:
        # weak stimulators represent that stimulant is not required for the reaction to occur
        stim_weak_str = '(Kd+'
        for i in range(rev_additive_stimulators):
            stim_weak_str += f'?W{i}*Kwr{i}+'
        
        rev_rate_law = stim_weak_str[:-1] + ')' + rev_rate_law[2:]

        # fill extra states
        extra_states = tuple([f'?W{i}' for i in range(rev_additive_stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kwr{i}' for i in range(rev_additive_stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kwr{i}': 1e-4 for i in range(rev_additive_stimulators)})

    if rev_allo_inhibitors > 0:
        # add the inhibitors to the equation, exact same thing as allo_stimulators but 
        # applied to the reverse rate law, with different parameter names 

        # add the stimulators to the equation
        altered_str = '*(1/(1+'
        for i in range(rev_allo_inhibitors):
            altered_str += f'?I{i}*Kir{i}+'

        rev_rate_law += altered_str[:-1] + '))'

        # fill extra states
        extra_states = tuple([f'?I{i}' for i in range(rev_allo_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kir{i}' for i in range(rev_allo_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kir{i}': 1e-4 for i in range(rev_allo_inhibitors)})

    if rev_comp_inhibitors > 0:
        # add the inhibitors to the equation, exact same thing as addi_stimulators but
        # applied to the reverse rate law, with different parameter names

        assert rev_additive_stimulators == 0, 'Currently this function does not support having both additive and competitive inhibitors'

        # add the stimulators to the equation
        stim_str = '(Kd/(1+'
        for i in range(rev_comp_inhibitors):
            stim_str += f'?C{i}*Kcr{i}+'

        rev_rate_law = stim_str[:-1] + '))' + rev_rate_law[2:]

        # fill extra states
        extra_states = tuple([f'?C{i}' for i in range(rev_comp_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kcr{i}' for i in range(rev_comp_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kcr{i}': 1e-4 for i in range(rev_comp_inhibitors)})

    return ReactionArchtype(
        'Mass Action General',
        reactants, products,
        parameters,
        forw_rate_law,
        extra_states=total_extra_states,
        assume_parameters_values=assume_parameters_values,
        assume_reactant_values={f'&R{i}': 100 for i in range(reactant_count)},
        assume_product_values={f'&P{i}': 0 for i in range(product_count)},
        reversible=True,
        reverse_rate_law=rev_rate_law)


def create_archtype_basal_michaelis(stimulators=0, stimulator_weak=0, allosteric_inhibitors=0, competitive_inhibitors=0):
    '''
    WARNING: basal currently only supports stimulator_weak and competitive_inhibitors
    The basal version do not generate Vmax as a parameter and only uses stimuator_weak and competitive_inhibitors
    '''
    if stimulators + allosteric_inhibitors + competitive_inhibitors + stimulator_weak == 0:
        return michaelis_menten

    # create the archtype

    archtype_name = 'Michaelis Menten General'

    reactants = ('&S',)
    products = ('&E',)
    upper_equation = 'Kc*&S'
    lower_equation = '(Km + &S)'
    total_extra_states = ()
    parameters = ('Km','Kc')
    assume_parameters_values={'Km': 100, 'Kc': 1}

    if stimulators > 0:
        # add the stimulators to the equation
        
        # first remove Kc from parameters
        parameters = tuple(p for p in parameters if p != "Kc")
        assume_parameters_values.pop("Kc", None)

        stim_str = '*('
        for i in range(stimulators):
            stim_str += f'&A{i}*Ka{i}+'
        
        upper_equation += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'&A{i}' for i in range(stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ka{i}' for i in range(stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ka{i}': 0.01 for i in range(stimulators)})

    if stimulator_weak > 0:
        # weak stimulators represent that stimulant is not required for the reaction to occur
        
        stim_weak_str = '(Kc+'
        for i in range(stimulator_weak):
            stim_weak_str += f"Kc{i}*&W{i}+"
        
        upper_equation = stim_weak_str[:-1] + ")" + "*&S"

        # fill extra states
        extra_states = tuple([f'&W{i}' for i in range(stimulator_weak)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kc{i}' for i in range(stimulator_weak)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kc{i}': 0.1 for i in range(stimulator_weak)})

    if allosteric_inhibitors > 0:
        # add the allosteric inhibitors to the equation
        inhb_allo_str = '/(1 + '
        for i in range(allosteric_inhibitors):
            inhb_allo_str += f'&L{i}/Ki{i} + '
        
        inhb_allo_str = inhb_allo_str[:-3] + ')'

        lower_equation += inhb_allo_str

        # fill extra states
        extra_states = tuple([f'&L{i}' for i in range(allosteric_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ki{i}' for i in range(allosteric_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ki{i}': 0.01 for i in range(allosteric_inhibitors)})
    
    if competitive_inhibitors > 0:
        # add the competitive inhibitors to the equation
        inhb_comp_str = '(Km*(1+'
        for i in range(competitive_inhibitors):
            inhb_comp_str += f'&I{i}*Kic{i}+'
        
        inhb_comp_str = inhb_comp_str[:-1] + ')'

        lower_equation = inhb_comp_str + lower_equation[3:]

        # fill extra states
        extra_states = tuple([f'&I{i}' for i in range(competitive_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kic{i}' for i in range(competitive_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kic{i}': 0.1 for i in range(competitive_inhibitors)})

    full_equation = f'{upper_equation}/{lower_equation}'

    general_reaction = ReactionArchtype(
        archtype_name,
        reactants, products,
        parameters,
        full_equation,
        extra_states=total_extra_states,
        assume_parameters_values=assume_parameters_values,
        assume_reactant_values={'&S': 100},
        assume_product_values={'&E': 0})
    
    return general_reaction

def create_archtype_michaelis_menten_v2(stimulators=0, stimulator_weak=0, allosteric_inhibitors=0, competitive_inhibitors=0): 
    '''
    WARNING: v2 currently only supports stimulator_weak and competitive_inhibitors
    The v2 version do not generate Vmax as a parameter and only uses stimuator_weak and competitive_inhibitors
    '''
    if stimulators + allosteric_inhibitors + competitive_inhibitors + stimulator_weak == 0:
        return michaelis_menten

    # create the archtype

    archtype_name = 'Michaelis Menten General'

    reactants = ('&S',)
    products = ('&E',)
    upper_equation = 'Kc*&S'
    lower_equation = '(Km + &S)'
    total_extra_states = ()
    parameters = ('Km','Kc')
    assume_parameters_values={'Km': 100, 'Kc': 1}

    if stimulators > 0:
        # add the stimulators to the equation
        
        # first remove Kc from parameters
        parameters = tuple(p for p in parameters if p != "Kc")
        assume_parameters_values.pop("Kc", None)

        stim_str = '*('
        for i in range(stimulators):
            stim_str += f'&A{i}*Ka{i}+'
        
        upper_equation += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'&A{i}' for i in range(stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ka{i}' for i in range(stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ka{i}': 0.01 for i in range(stimulators)})

    if stimulator_weak > 0:
        # weak stimulators represent that stimulant is not required for the reaction to occur
        
        # first remove Kc from parameters
        parameters = tuple(p for p in parameters if p != 'Kc')
        assume_parameters_values.pop('Kc', None)
        
        stim_weak_str = '('
        for i in range(stimulator_weak):
            stim_weak_str += f"Kc{i}*&W{i}+"
        
        upper_equation = stim_weak_str[:-1] + ")" + "*&S"

        # fill extra states
        extra_states = tuple([f'&W{i}' for i in range(stimulator_weak)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kc{i}' for i in range(stimulator_weak)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kc{i}': 0.1 for i in range(stimulator_weak)})

    if allosteric_inhibitors > 0:
        # add the allosteric inhibitors to the equation
        inhb_allo_str = '/(1 + '
        for i in range(allosteric_inhibitors):
            inhb_allo_str += f'&L{i}/Ki{i} + '
        
        inhb_allo_str = inhb_allo_str[:-3] + ')'

        lower_equation += inhb_allo_str

        # fill extra states
        extra_states = tuple([f'&L{i}' for i in range(allosteric_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ki{i}' for i in range(allosteric_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ki{i}': 0.01 for i in range(allosteric_inhibitors)})
    
    if competitive_inhibitors > 0:
        # add the competitive inhibitors to the equation
        inhb_comp_str = '(Km*(1+'
        for i in range(competitive_inhibitors):
            inhb_comp_str += f'&I{i}*Kic{i}+'
        
        inhb_comp_str = inhb_comp_str[:-1] + ')'

        lower_equation = inhb_comp_str + lower_equation[3:]

        # fill extra states
        extra_states = tuple([f'&I{i}' for i in range(competitive_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kic{i}' for i in range(competitive_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kic{i}': 0.1 for i in range(competitive_inhibitors)})

    full_equation = f'{upper_equation}/{lower_equation}'

    general_reaction = ReactionArchtype(
        archtype_name,
        reactants, products,
        parameters,
        full_equation,
        extra_states=total_extra_states,
        assume_parameters_values=assume_parameters_values,
        assume_reactant_values={'&S': 100},
        assume_product_values={'&E': 0})
    
    return general_reaction


def create_archtype_michaelis_menten(stimulators=0, stimulator_weak=0, allosteric_inhibitors=0, competitive_inhibitors=0):

    if stimulators + allosteric_inhibitors + competitive_inhibitors + stimulator_weak == 0:
        return michaelis_menten

    # create the archtype

    archtype_name = 'Michaelis Menten General'

    reactants = ('&S',)
    products = ('&E',)
    upper_equation = 'Vmax*&S'
    lower_equation = '(Km + &S)'
    total_extra_states = ()
    parameters = ('Km', 'Vmax')
    assume_parameters_values={'Km': 100, 'Vmax': 1}

    if stimulators > 0:
        # add the stimulators to the equation
        stim_str = '*('
        for i in range(stimulators):
            stim_str += f'&A{i}*Ka{i}+'
        
        upper_equation += stim_str[:-1] + ')'

        # fill extra states
        extra_states = tuple([f'&A{i}' for i in range(stimulators)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Ka{i}' for i in range(stimulators)])

        # fill assume parameters values
        assume_parameters_values.update({f'Ka{i}': 0.01 for i in range(stimulators)})

    if stimulator_weak > 0:
        # weak stimulators represent that stimulant is not required for the reaction to occur
        stim_weak_str = '(Vmax+'
        for i in range(stimulator_weak):
            stim_weak_str += f'&W{i}*Kw{i}+'
        
        upper_equation = stim_weak_str[:-1] + ')' + upper_equation[4:]

        # fill extra states
        extra_states = tuple([f'&W{i}' for i in range(stimulator_weak)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kw{i}' for i in range(stimulator_weak)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kw{i}': 0.1 for i in range(stimulator_weak)})

    if allosteric_inhibitors > 0:
        # add the allosteric inhibitors to the equation
        inhb_allo_str = '*(1+'
        for i in range(allosteric_inhibitors):
            inhb_allo_str += f'&L{i}*Kil{i}+'
        
        inhb_allo_str = inhb_allo_str[:-1] + ')'

        lower_equation += inhb_allo_str

        # fill extra states
        extra_states = tuple([f'&L{i}' for i in range(allosteric_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kil{i}' for i in range(allosteric_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kil{i}': 0.01 for i in range(allosteric_inhibitors)})
    
    if competitive_inhibitors > 0:
        # add the competitive inhibitors to the equation
        inhb_comp_str = '(Km*(1+'
        for i in range(competitive_inhibitors):
            inhb_comp_str += f'&I{i}*Kic{i}+'
        
        inhb_comp_str = inhb_comp_str[:-1] + ')'

        lower_equation = inhb_comp_str + lower_equation[3:]

        # fill extra states
        extra_states = tuple([f'&I{i}' for i in range(competitive_inhibitors)])
        total_extra_states += extra_states

        # fill parameters
        parameters += tuple([f'Kic{i}' for i in range(competitive_inhibitors)])

        # fill assume parameters values
        assume_parameters_values.update({f'Kic{i}': 0.1 for i in range(competitive_inhibitors)})

    full_equation = f'{upper_equation}/{lower_equation}'

    general_reaction = ReactionArchtype(
        archtype_name,
        reactants, products,
        parameters,
        full_equation,
        extra_states=total_extra_states,
        assume_parameters_values=assume_parameters_values,
        assume_reactant_values={'&S': 100},
        assume_product_values={'&E': 0})

    return general_reaction



# HTTP Solver
# This class sends HTTP requests to an external server for ODE simulation

from typing import Dict, Optional
import pandas as pd
import requests
from .Solver import Solver


class HTTPSolver(Solver):
    """
    A solver that sends HTTP POST requests to an external server for ODE simulation.

    The server receives simulation parameters (start, stop, step) and optional
    state/parameter overrides via JSON, and returns results as JSON.

    Server Implementation Guide:
        See docs/httpsolver_api.md for complete API specification including:
        - Request/response formats
        - Error handling
        - Example server implementations (FastAPI, Node.js/Express)

    Example:
        solver = HTTPSolver()
        solver.compile("http://localhost:8000/simulate")
        solver.set_state_values({"S1": 100.0})
        results = solver.simulate(0, 1000, 100)
    """

    def __init__(self):
        super().__init__()
        self.endpoint: Optional[str] = None
        self.timeout: float = 300.0  # Default timeout for requests (seconds)
        self.state_values: Dict[str, float] = {}
        self.param_values: Dict[str, float] = {}

    def compile(self, compile_str: str, **kwargs) -> bool:
        """
        Compile the model by storing the HTTP endpoint URL.

        Args:
            compile_str: The HTTP endpoint URL (e.g., "http://localhost:8000/simulate")
            **kwargs: Additional arguments:
                - timeout: Request timeout in seconds (default: 300.0)
                - headers: Dict of HTTP headers to include in requests
                - auth: Tuple of (username, password) for basic auth

        Returns:
            True
        """
        self.endpoint = compile_str.rstrip('/')  # Remove trailing slash if present
        self.timeout = kwargs.get('timeout', 300.0)
        self.headers = kwargs.get('headers', {})
        self.auth = kwargs.get('auth')
        
        # perform a simple validation of the URL
        try:
            response = requests.head(self.endpoint, timeout=self.timeout)
            response.raise_for_status()
        except requests.RequestException as e:
            raise ValueError(f"Could not reach endpoint '{self.endpoint}': {e}")

        return True

    def simulate(self, start: float, stop: float, step: float) -> pd.DataFrame:
        """
        Simulate by sending a POST request to the HTTP endpoint.

        Args:
            start: Start time of the simulation
            stop: Stop time of the simulation
            step: Step size for the simulation

        Returns:
            pandas DataFrame with 'time' column and species columns

        Raises:
            requests.RequestException: If the HTTP request fails
            ValueError: If the response is invalid
        """
        if self.endpoint is None:
            raise RuntimeError("Solver not compiled. Call compile() with an endpoint URL first.")

        # Prepare JSON payload
        payload = {
            'start': start,
            'stop': stop,
            'step': step
        }

        # Include state and parameter values if they have been set
        if self.state_values:
            payload['state_values'] = self.state_values
        if self.param_values:
            payload['parameter_values'] = self.param_values

        # Prepare headers
        headers = self.headers.copy()
        headers['Content-Type'] = 'application/json'

        # Send POST request
        response = requests.post(
            self.endpoint,
            json=payload,
            timeout=self.timeout,
            headers=headers,
            auth=self.auth
        )
        response.raise_for_status()

        # Clear state and parameter values after sending (hot swappability)
        self.state_values.clear()
        self.param_values.clear()

        # Parse JSON response
        try:
            data = response.json()
        except ValueError as e:
            raise ValueError(f"Invalid JSON response from server: {e}")

        # Convert to DataFrame
        try:
            df = pd.DataFrame(data)
            self.last_sim_result = df
            return df
        except Exception as e:
            raise ValueError(f"Could not convert response to DataFrame: {e}")

    def set_state_values(self, state_values: Dict[str, float]) -> bool:
        """
        Set state values to be sent in the next simulate() call.

        The values replace any previously set state values and are included
        in the payload of the next simulate() request. Values are cleared
        after the request completes, enabling hot-swappability.

        Args:
            state_values: Dictionary mapping state names to their values

        Returns:
            True if the state values were set successfully
        """
        self.state_values = state_values.copy()
        return True

    def set_parameter_values(self, parameter_values: Dict[str, float]) -> bool:
        """
        Set parameter values to be sent in the next simulate() call.

        The values replace any previously set parameter values and are included
        in the payload of the next simulate() request. Values are cleared
        after the request completes, enabling hot-swappability.

        Args:
            parameter_values: Dictionary mapping parameter names to their values

        Returns:
            True if the parameter values were set successfully
        """
        self.param_values = parameter_values.copy()
        return True

    def _get_base_url(self) -> str:
        """Get the base URL from the endpoint (remove /simulate suffix if present)."""
        if self.endpoint is None:
            raise RuntimeError("Solver not compiled. Call compile() with an endpoint URL first.")

        # Remove /simulate suffix if present to get base URL
        base_url = self.endpoint
        if base_url.endswith('/simulate'):
            base_url = base_url[:-9]  # Remove '/simulate'
        return base_url.rstrip('/')

    def get_state_defaults(self) -> Dict[str, float]:
        """
        Fetch default initial values for all states from the HTTP endpoint.

        Sends a GET request to the /states endpoint to retrieve the server's
        default state values (species concentrations).

        Returns:
            Dictionary mapping state names to their default initial values

        Raises:
            RuntimeError: If solver not compiled
            requests.RequestException: If the HTTP request fails
            ValueError: If the response is invalid JSON

        Example:
            >>> solver = HTTPSolver()
            >>> solver.compile("http://localhost:8000/simulate")
            >>> states = solver.get_state_defaults()
            >>> print(states)
            {'Ribosome': 1900.0, 'ppERK': 11.1, 'ppAKT': 0.389, ...}
        """
        base_url = self._get_base_url()
        states_url = f"{base_url}/states"

        response = requests.get(
            states_url,
            timeout=self.timeout,
            headers=self.headers,
            auth=self.auth
        )
        response.raise_for_status()

        try:
            return response.json()
        except ValueError as e:
            raise ValueError(f"Invalid JSON response from /states endpoint: {e}")

    def get_parameter_defaults(self) -> Dict[str, float]:
        """
        Fetch default values for all parameters from the HTTP endpoint.

        Sends a GET request to the /parameters endpoint to retrieve the server's
        default parameter values.

        Returns:
            Dictionary mapping parameter names to their default values

        Raises:
            RuntimeError: If solver not compiled
            requests.RequestException: If the HTTP request fails
            ValueError: If the response is invalid JSON

        Example:
            >>> solver = HTTPSolver()
            >>> solver.compile("http://localhost:8000/simulate")
            >>> params = solver.get_parameter_defaults()
            >>> print(params)
            {'k1_1': 0.0042005, 'k3_1': 140.7475, ...}
        """
        base_url = self._get_base_url()
        params_url = f"{base_url}/parameters"

        response = requests.get(
            params_url,
            timeout=self.timeout,
            headers=self.headers,
            auth=self.auth
        )
        response.raise_for_status()

        try:
            return response.json()
        except ValueError as e:
            raise ValueError(f"Invalid JSON response from /parameters endpoint: {e}")

    def get_species_list(self):
        """
        Get list of species (state variable) names in the compiled model.

        Fetches species names from the HTTP server's /states endpoint
        without running a simulation.

        Returns:
            List of species names

        Raises:
            RuntimeError: If solver not compiled
            requests.RequestException: If the HTTP request fails
        """
        return list(self.get_state_defaults().keys())


# abstract class solver 
# This class is used to solve the problem

from abc import ABC, abstractmethod
from typing import List, Dict, Any, Tuple
import pandas as pd

class Solver(ABC):
    """
    Abstract class for an ODE solver that can run simulations 
    """
    
    def __init__(self):
        super().__init__()
        self.last_sim_result = None
        
    @abstractmethod
    def compile(self, compile_str: str, **kwargs):
        """
        Return a model instance which can be used to run simulations. 
        The compile_str is a string either in antimony or sbml format, alternatively it can be a file path to a file in either of these formats.
        The kwargs are additional arguments that can be passed for the creation of the model instance.
        """
        pass 

    @abstractmethod
    def simulate(self, start: float, stop: float, step: float) -> pd.DataFrame:
        """
        Simulate the problem from start to stop with a given step size.
        Returns a pandas dataframe with the results with the following columns:
        - time: time points of the simulation
        - [species]: species names
        with each row corresponds to a time point and each column corresponds to a species.
        """
        pass
    
    def set_state_values(self, state_values: Dict[str, float]) -> bool:
        """
        Hot swapping of state variables in the running instance of the model.
        Set the values of state variables in the model instance, this should only possible after compiling the model. 
        Not every solver will support this, so it is possible that this function to return an not implemented error.
        returns True if the state variable was set successfully, False otherwise.
        """
        pass
     
    def set_parameter_values(self, parameter_values: Dict[str, float]) -> bool:
        """
        Hot swapping of parameters in the running instance of the model.
        Set the values of parameter variables in the model instance, this should only possible after compiling the model.
        Not every solver will support this, so it is possible that this function to return an not implemented error.
        returns True if the state variable was set successfully, False otherwise.
        """
        pass

    def get_species_list(self) -> List[str]:
        """
        Get list of species (state variable) names in the compiled model.

        This method retrieves species names without running a simulation,
        which is useful for discovery and validation purposes.

        Returns:
            List of species names (state variables)

        Raises:
            RuntimeError: If solver not compiled
        """
        raise NotImplementedError("Subclasses must implement get_species_list()")



"""
Parameter mapping utilities for ModelBuilder objects.
Provides functions to analyze parameter-reaction relationships.
"""

from typing import Dict, List, Optional, Tuple, Any
import re
from ..ModelBuilder import ModelBuilder


def get_parameter_reaction_map(model_builder: ModelBuilder) -> Dict[str, Dict]:
    """
    Map each parameter to its associated reaction details.
    
    Args:
        model_builder: The ModelBuilder object to analyze
        
    Returns:
        Dictionary mapping parameter names to reaction details:
        {
            'Km_J0': {
                'reaction_index': 0,
                'parameter_type': 'Km',
                'reactants': ['O'],
                'products': ['Oa'],
                'reaction_description': 'O -> Oa'
            },
            ...
        }
    """
    if not model_builder.pre_compiled:
        raise ValueError("Model must be pre-compiled before parameter mapping")
    
    param_map = {}
    all_params = model_builder.get_parameters()
    
    for param_name in all_params.keys():
        # Parse parameter name: {type}_{index}
        match = re.match(r'^([a-zA-Z][a-zA-Z0-9]*)_(J\d+)$', param_name)
        if not match:
            # Try alternative pattern for linked parameters or custom names
            match = re.match(r'^([a-zA-Z][a-zA-Z0-9]*)_(.+)$', param_name)
            if not match:
                continue
        
        param_type = match.group(1)
        reaction_id = match.group(2)
        
        # Extract reaction index from J0, J1, etc.
        reaction_idx = None
        if reaction_id.startswith('J'):
            try:
                reaction_idx = int(reaction_id[1:])
            except ValueError:
                reaction_idx = None
        
        reaction_info = {
            'parameter_name': param_name,
            'parameter_type': param_type,
            'reaction_id': reaction_id,
            'reaction_index': reaction_idx,
            'reactants': [],
            'products': [],
            'reaction_description': 'Unknown'
        }
        
        # Try to get reaction details if we have an index
        if reaction_idx is not None and reaction_idx < len(model_builder.reactions):
            reaction = model_builder.reactions[reaction_idx]
            reaction_info['reactants'] = list(reaction.reactants_names)
            reaction_info['products'] = list(reaction.products_names)
            reaction_info['reaction_description'] = (
                f"{' + '.join(reaction.reactants_names)} -> "
                f"{' + '.join(reaction.products_names)}"
            )
            
            # Try to get archtype information if available
            try:
                reaction_info['parameter_roles'] = list(reaction.archtype.parameters)
            except AttributeError:
                reaction_info['parameter_roles'] = []
        
        param_map[param_name] = reaction_info
    
    return param_map


def find_parameter_by_role(
    model_builder: ModelBuilder, 
    role: str, 
    state_var: Optional[str] = None
) -> List[str]:
    """
    Find parameters that match a specific role.
    
    Args:
        model_builder: The ModelBuilder object to analyze
        role: Parameter role to search for (e.g., 'Vmax', 'Km', 'Kc')
               Use None to match all parameters (filtered by state_var only)
        state_var: Optional state variable to filter by (e.g., 'O', 'Oa')
                    Checks if state_var is involved in the reaction
        
    Returns:
        List of parameter names matching the criteria
    """
    if not model_builder.pre_compiled:
        raise ValueError("Model must be pre-compiled before parameter search")
    
    param_map = get_parameter_reaction_map(model_builder)
    matching_params = []
    
    for param_name, param_info in param_map.items():
        # Filter by role if specified
        if role is not None:
            param_type = param_info.get('parameter_type', '')
            # Case-insensitive partial matching
            if role.lower() not in param_type.lower():
                continue
        
        # Filter by state variable if specified
        if state_var is not None:
            reactants = param_info.get('reactants', [])
            products = param_info.get('products', [])
            
            # Check exact match
            is_involved = (state_var in reactants) or (state_var in products)
            
            # Check for activated/deactivated forms
            if not is_involved:
                # Handle X vs Xa patterns
                if state_var.endswith('a'):
                    base_var = state_var[:-1]  # Remove 'a'
                    is_involved = (base_var in reactants) or (base_var in products)
                else:
                    activated_var = state_var + 'a'
                    is_involved = (activated_var in reactants) or (activated_var in products)
            
            if not is_involved:
                continue
        
        matching_params.append(param_name)
    
    return matching_params


def explain_reaction_parameters(
    model_builder: ModelBuilder, 
    reaction_index: int
) -> str:
    """
    Generate human-readable explanation of all parameters in a reaction.
    
    Args:
        model_builder: The ModelBuilder object
        reaction_index: Index of the reaction to explain (0-based)
        
    Returns:
        Human-readable explanation string
    """
    if not model_builder.pre_compiled:
        raise ValueError("Model must be pre-compiled")
    
    if reaction_index >= len(model_builder.reactions):
        raise IndexError(f"Reaction index {reaction_index} out of bounds")
    
    reaction = model_builder.reactions[reaction_index]
    param_map = get_parameter_reaction_map(model_builder)
    
    # Find parameters for this reaction
    reaction_params = []
    for param_name, param_info in param_map.items():
        if param_info.get('reaction_index') == reaction_index:
            reaction_params.append((param_name, param_info))
    
    if not reaction_params:
        return f"Reaction {reaction_index} has no directly associated parameters"
    
    # Build explanation
    reactants = reaction.reactants_names
    products = reaction.products_names
    explanation = [f"Reaction {reaction_index}: {', '.join(reactants)} → {', '.join(products)}"]
    explanation.append("Parameters:")
    
    for param_name, param_info in reaction_params:
        param_type = param_info.get('parameter_type', 'Unknown')
        role_desc = _describe_parameter_role(param_type, reactants, products)
        explanation.append(f"  - {param_name}: {role_desc}")
    
    return "\n".join(explanation)


def _describe_parameter_role(
    param_type: str, 
    reactants: Tuple[str, ...], 
    products: Tuple[str, ...]
) -> str:
    """
    Generate human-readable description of a parameter's role.
    
    Args:
        param_type: Parameter type (e.g., 'Vmax', 'Km', 'Kc')
        reactants: Reactant names
        products: Product names
        
    Returns:
        Description string
    """
    param_type_lower = param_type.lower()
    
    # Common parameter roles based on naming patterns
    if 'vmax' in param_type_lower:
        if reactants and products:
            return f"Maximum rate for conversion of {reactants[0]} to {products[0]}"
        return "Maximum reaction rate"
    
    elif 'km' in param_type_lower:
        if reactants:
            return f"Michaelis constant for substrate {reactants[0]}"
        return "Michaelis constant (substrate affinity)"
    
    elif 'kc' in param_type_lower:
        return "Constitutive/basal rate constant"
    
    elif param_type_lower.startswith('ka'):
        return "Allosteric activation constant"
    
    elif param_type_lower.startswith('ki'):
        if 'c' in param_type_lower:
            return "Competitive inhibition constant"
        return "Allosteric inhibition constant"
    
    elif param_type_lower.startswith('k'):
        return "Rate constant"
    
    else:
        return f"Parameter of type {param_type}"


def get_parameters_for_state(
    model_builder: ModelBuilder, 
    state_var: str
) -> Dict[str, List[str]]:
    """
    Get all parameters affecting a specific state variable.
    
    Args:
        model_builder: The ModelBuilder object
        state_var: State variable name (e.g., 'O', 'R1a')
        
    Returns:
        Dictionary with lists of parameters:
        {
            'as_reactant': ['Km_J0', 'Vmax_J0'],  # Parameters in reactions where state is reactant
            'as_product': ['Kc_J1'],              # Parameters in reactions where state is product
            'all': ['Km_J0', 'Vmax_J0', 'Kc_J1']  # All parameters affecting state
        }
        
    Raises:
        ValueError: If state_var is not found in the model
    """
    # Check if state exists in the model
    all_states = model_builder.get_state_variables()
    if state_var not in all_states:
        # Check for activated/deactivated forms
        if state_var.endswith('a'):
            base_var = state_var[:-1]
            if base_var not in all_states:
                raise ValueError(f"State variable '{state_var}' not found in the model")
        else:
            activated_var = state_var + 'a'
            if activated_var not in all_states:
                raise ValueError(f"State variable '{state_var}' not found in the model")
    
    param_map = get_parameter_reaction_map(model_builder)
    
    as_reactant = []
    as_product = []
    
    for param_name, param_info in param_map.items():
        reactants = param_info.get('reactants', [])
        products = param_info.get('products', [])
        
        if state_var in reactants:
            as_reactant.append(param_name)
        elif state_var in products:
            as_product.append(param_name)
        else:
            # Check for activated/deactivated forms
            if state_var.endswith('a'):
                base_var = state_var[:-1]
                if base_var in reactants or base_var in products:
                    as_reactant.append(param_name)
            else:
                activated_var = state_var + 'a'
                if activated_var in reactants or activated_var in products:
                    as_product.append(param_name)
    
    all_params = list(set(as_reactant + as_product))
    
    return {
        'as_reactant': as_reactant,
        'as_product': as_product,
        'all': all_params
    }



"""
High-level API for Synthetic library.

This module provides a simple, intuitive interface for generating synthetic drug
response datasets compatible with scikit-learn's `make_regression` format.

The API uses DegreeInteractionSpec for network topology and KineticParameterTuner
for biologically plausible kinetic parameters.

Example usage:
    from synthetic import Builder, make_dataset_drug_response

    # Create a virtual cell model
    vc = Builder.specify(degree_cascades=[1, 2, 5])

    # Generate sklearn-compatible dataset
    X, y = make_dataset_drug_response(n=1000, cell_model=vc, target_specie='Oa')
"""

from typing import List, Dict, Tuple, Optional, Any, Union
import numpy as np
import pandas as pd

from .Specs.DegreeInteractionSpec import DegreeInteractionSpec
from .Specs.Drug import Drug
from .ModelBuilder import ModelBuilder
from .Solver.ScipySolver import ScipySolver
from .Solver.RoadrunnerSolver import RoadrunnerSolver
from .utils.kinetic_tuner import KineticParameterTuner


class VirtualCell:
    """
    High-level abstraction for a virtual cell model.

    Uses DegreeInteractionSpec for network topology and KineticParameterTuner
    for biologically plausible kinetic parameters.

    This class provides a simple API for creating and managing virtual cell
    models without needing to understand the underlying complexity of the
    Synthetic library's architecture.
    """

    def __init__(
        self,
        degree_cascades: List[int],
        name: str = "VirtualCell",
        random_seed: Optional[int] = None,
        feedback_density: float = 0.5,
        auto_compile: bool = True,
        auto_drug: bool = True,
        drug_name: str = "D",
        drug_start_time: float = 5000.0,
        drug_value: float = 100.0,
        drug_regulation_type: str = "down",
        simulation_end: float = 10000.0,
    ):
        """
        Initialize a virtual cell model.

        Args:
            degree_cascades: List of cascade counts per degree, e.g., [1, 2, 5]
            name: Name for the virtual cell
            random_seed: Optional random seed for reproducibility
            feedback_density: Proportion of feedback connections (0-1)
            auto_compile: Compile immediately after creation (default: True)
            auto_drug: Auto-generate a drug targeting degree 1 R species (default: True)
            drug_name: Name for auto-generated drug (default: "D")
            drug_start_time: Time at which auto-drug becomes active (default: 5000.0)
            drug_value: Active concentration for auto-drug (default: 100.0)
            drug_regulation_type: Regulation type for auto-drug: "up" or "down" (default: "down")
            simulation_end: Simulation end time for make_dataset_drug_response (default: 10000.0)
        """
        self._degree_cascades = degree_cascades
        self._name = name
        self._random_seed = random_seed
        self._feedback_density = feedback_density
        self._auto_compile = auto_compile
        self._auto_drug = auto_drug
        self._auto_drug_name: Optional[str] = drug_name if auto_drug else None
        self._drug_start_time = drug_start_time
        self._auto_drug_value = drug_value
        self._auto_drug_regulation_type = drug_regulation_type
        self._simulation_end = simulation_end
        self._drugs: List[Tuple[Drug, Optional[float]]] = []
        self._spec: Optional[DegreeInteractionSpec] = None
        self._model: Optional[ModelBuilder] = None
        self._tuner: Optional[KineticParameterTuner] = None
        self._compiled = False

    def add_drug(
        self,
        name: str,
        start_time: float = 500.0,
        default_value: float = 0.0,
        regulation: Optional[List[str]] = None,
        regulation_type: Optional[List[str]] = None,
        value: Optional[float] = None,
    ) -> 'VirtualCell':
        """
        Add a drug to the cell model.

        Drugs can only target degree 1 R species. The drug appears at start_time
        via a piecewise assignment rule and regulates the specified species.

        Args:
            name: Name of the drug
            start_time: Time at which drug becomes active (default: 500.0)
            default_value: Default value when drug is not active (default: 0.0)
            regulation: List of target species (must be degree 1 R species)
            regulation_type: List of regulation types ('up' or 'down')
            value: Optional override value for when drug is active

        Returns:
            self for method chaining
        """
        drug = Drug(
            name=name,
            start_time=start_time,
            default_value=default_value,
            regulation=regulation,
            regulation_type=regulation_type,
        )
        self._drugs.append((drug, value))
        return self

    def list_drugs(self) -> List[Dict[str, Any]]:
        """
        List all drugs in the system (both auto-generated and manually added).

        Returns:
            List of dictionaries with drug information:
            [{'name': 'D', 'targets': ['R1_1'], 'types': ['down'],
              'start_time': 5000.0, 'default_value': 0.0, 'is_auto': True}, ...]
        """
        drugs_info = []
        for drug, value in self._drugs:
            drugs_info.append({
                'name': drug.name,
                'targets': drug.regulation,
                'types': drug.regulation_type,
                'start_time': drug.start_time,
                'default_value': drug.default_value,
                'is_auto': drug.name == self._auto_drug_name
            })
        return drugs_info

    def _generate_auto_drug(self) -> None:
        """
        Generate auto-drug based on degree_cascades[0].

        Creates a drug targeting all degree 1 R species (R1_1, R1_2, etc.)
        with count based on degree_cascades[0].
        """
        cascade_count = self._degree_cascades[0]
        targets = []

        # Generate targets: R1_1, R1_2, ... based on cascade_count
        for i in range(cascade_count):
            targets.append(f"R1_{i + 1}")

        # Create and store auto-drug
        drug = Drug(
            name=self._auto_drug_name,
            start_time=self._drug_start_time,
            default_value=0.0,
            regulation=targets,
            regulation_type=[self._auto_drug_regulation_type] * len(targets),
        )
        self._drugs.append((drug, self._auto_drug_value))

    def compile(
        self,
        mean_range_species: Tuple[int, int] = (50, 150),
        rangeScale_params: Tuple[float, float] = (0.8, 1.2),
        rangeMultiplier_params: Tuple[float, float] = (0.9, 1.1),
        use_kinetic_tuner: bool = True,
        active_percentage_range: Tuple[float, float] = (0.3, 0.7),
        X_total_multiplier: float = 5.0,
        ki_val: float = 100.0,
        v_max_f_random_range: Tuple[float, float] = (5.0, 10.0),
    ) -> 'VirtualCell':
        """
        Compile the model (lazy initialization).

        Creates the underlying DegreeInteractionSpec, ModelBuilder, and optionally
        applies KineticParameterTuner for biologically plausible parameters.

        If auto_drug is enabled, automatically generates a drug targeting degree 1
        R species and applies its parameters to the model.

        Args:
            mean_range_species: Range for initial species values
            rangeScale_params: Range for parameter scaling
            rangeMultiplier_params: Range for parameter multiplier
            use_kinetic_tuner: Whether to use KineticParameterTuner (default: True)
            active_percentage_range: Target active percentage range for tuning
            X_total_multiplier: Multiplier for Km_b calculation
            ki_val: Constant Ki value for inhibitors
            v_max_f_random_range: Range for total forward Vmax

        Returns:
            self for method chaining
        """
        # Generate auto-drug if enabled
        if self._auto_drug:
            self._generate_auto_drug()

        # Create specification
        self._spec = DegreeInteractionSpec(degree_cascades=self._degree_cascades)
        self._spec.generate_specifications(
            random_seed=self._random_seed,
            feedback_density=self._feedback_density,
        )

        # Add drugs (auto-drug + any manually added)
        for drug, value in self._drugs:
            self._spec.add_drug(drug, value)

        # Generate model
        self._model = self._spec.generate_network(
            network_name=self._name,
            mean_range_species=mean_range_species,
            rangeScale_params=rangeScale_params,
            rangeMultiplier_params=rangeMultiplier_params,
            random_seed=self._random_seed,
        )
        self._model.precompile()

        # Apply kinetic tuning if requested
        if use_kinetic_tuner:
            self._tuner = KineticParameterTuner(self._model, random_seed=self._random_seed)
            updated_params = self._tuner.generate_parameters(
                active_percentage_range=active_percentage_range,
                X_total_multiplier=X_total_multiplier,
                ki_val=ki_val,
                v_max_f_random_range=v_max_f_random_range,
            )
            for param_name, value in updated_params.items():
                self._model.set_parameter(param_name, value)

        # Apply auto-drug parameters using regulator_parameter_map
        if self._auto_drug_name:
            regulator_parameter_map = self._model.get_regulator_parameter_map()
            drug_params = regulator_parameter_map.get(self._auto_drug_name, {})
            for param_name in drug_params:
                self._model.set_parameter(param_name, self._auto_drug_value)

        self._compiled = True
        return self

    def get_species_names(self, exclude_drugs: bool = True) -> List[str]:
        """
        Get list of species names.

        Args:
            exclude_drugs: Whether to exclude drug species (default: True)

        Returns:
            List of species names
        """
        if not self._compiled:
            raise ValueError("Model must be compiled first. Call compile().")
        return self._model.get_state_variables().keys()

    def get_initial_values(self, exclude_drugs: bool = True) -> Dict[str, float]:
        """
        Get initial species values from the compiled model.

        Args:
            exclude_drugs: Whether to exclude drug species (default: True)

        Returns:
            Dictionary mapping species names to initial values
        """
        if not self._compiled:
            raise ValueError("Model must be compiled first. Call compile().")

        initial_values = self._model.get_state_variables().copy()

        if exclude_drugs:
            # Remove drugs from initial values (they are in model.variables)
            for drug_name in [d[0].name for d in self._drugs]:
                initial_values.pop(drug_name, None)

        return initial_values

    def get_target_concentrations(self) -> Dict[str, float]:
        """
        Get target active concentrations from the kinetic tuner.

        Returns an empty dict if kinetic tuning was not used during compilation.

        Returns:
            Dictionary mapping species names to target concentrations
        """
        if not self._compiled:
            raise ValueError("Model must be compiled first. Call compile().")
        if self._tuner is None:
            return {}
        return self._tuner.get_target_concentrations()

    @property
    def spec(self) -> DegreeInteractionSpec:
        """
        Access to underlying specification.

        Returns the DegreeInteractionSpec object. Raises ValueError if model
        has not been compiled.

        Returns:
            DegreeInteractionSpec object
        """
        if self._spec is None:
            raise ValueError("Model must be compiled first. Call compile().")
        return self._spec

    @property
    def model(self) -> ModelBuilder:
        """
        Access to underlying model.

        Returns the ModelBuilder object. Raises ValueError if model
        has not been compiled.

        Returns:
            ModelBuilder object
        """
        if self._model is None:
            raise ValueError("Model must be compiled first. Call compile().")
        return self._model

    @property
    def tuner(self) -> KineticParameterTuner:
        """
        Access to underlying kinetic tuner.

        Returns the KineticParameterTuner object, or None if kinetic
        tuning was not used during compilation.

        Returns:
            KineticParameterTuner object or None
        """
        return self._tuner


class Builder:
    """
    Factory for creating virtual cell models.

    Provides a static method for creating VirtualCell instances with
    a clean, fluent API.
    """

    @staticmethod
    def specify(
        degree_cascades: List[int],
        name: str = "VirtualCell",
        random_seed: Optional[int] = None,
        feedback_density: float = 1,
        auto_compile: bool = True,
        auto_drug: bool = True,
        drug_name: str = "D",
        drug_start_time: float = 5000.0,
        drug_value: float = 100.0,
        drug_regulation_type: str = "down",
        simulation_end: float = 10000.0,
    ) -> VirtualCell:
        """
        Create a virtual cell specification.

        Args:
            degree_cascades: List of cascade counts per degree, e.g., [1, 2, 5]
            name: Name for the virtual cell
            random_seed: Optional random seed for reproducibility
            feedback_density: Proportion of feedback connections (0-1)
            auto_compile: Compile immediately after creation (default: True)
            auto_drug: Auto-generate a drug targeting degree 1 R species (default: True)
            drug_name: Name for auto-generated drug (default: "D")
            drug_start_time: Time at which auto-drug becomes active (default: 5000.0)
            drug_value: Active concentration for auto-drug (default: 100.0)
            drug_regulation_type: Regulation type for auto-drug: "up" or "down" (default: "down")
            simulation_end: Simulation end time for make_dataset_drug_response (default: 10000.0)

        Returns:
            VirtualCell instance (compiled if auto_compile=True)
        """
        vc = VirtualCell(
            degree_cascades=degree_cascades,
            name=name,
            random_seed=random_seed,
            feedback_density=feedback_density,
            auto_compile=auto_compile,
            auto_drug=auto_drug,
            drug_name=drug_name,
            drug_start_time=drug_start_time,
            drug_value=drug_value,
            drug_regulation_type=drug_regulation_type,
            simulation_end=simulation_end,
        )
        if auto_compile:
            vc.compile()
        return vc


def make_dataset_drug_response(
    n: int,
    cell_model: VirtualCell,
    target_specie: str = 'Oa',
    perturbation_type: str = 'conserve_rules',
    perturbation_params: Optional[Dict[str, Any]] = None,
    parameter_values: Optional[Dict[str, float]] = None,
    param_perturbation_type: str = 'lognormal',
    param_perturbation_params: Optional[Dict[str, Any]] = None,
    simulation_params: Optional[Dict[str, Any]] = None,
    seed: Optional[int] = None,
    param_seed: Optional[int] = None,
    solver_type: str = 'roadrunner',
    jit: bool = True,
    verbose: bool = False,
    n_cores: int = 1,
    require_all_successful: bool = False,
    return_details: bool = True,
    capture_all_species: bool = True,
    exclude_outcome_from_features: bool = True,
    exclude_activated_from_features: bool = True,
    as_pandas: bool = True,
    return_timecourse: bool = False,
) -> Union[Tuple[pd.DataFrame, pd.Series], Tuple[np.ndarray, np.ndarray], Dict[str, Any]]:
    """
    Generate synthetic drug response dataset.

    Creates a dataset compatible with scikit-learn's make_regression format.
    The feature matrix X contains species concentrations, and the target
    vector y contains the outcome values.

    Drugs are NOT included in the X features - they affect the simulation
    but are not part of the feature matrix.

    Args:
        n: Number of samples to generate
        cell_model: VirtualCell instance (must be compiled)
        target_specie: Name of the outcome species to use as target (default: 'Oa')
        perturbation_type: Type of initial value perturbation ('uniform', 'gaussian', 'lognormal', 'lhs')
        perturbation_params: Parameters for perturbation distribution
        parameter_values: Dictionary of kinetic parameter values to perturb (optional)
        param_perturbation_type: Type of kinetic parameter perturbation ('none', 'uniform', 'gaussian', 'lognormal', 'lhs')
        param_perturbation_params: Parameters for kinetic parameter perturbation (optional)
        simulation_params: Dictionary with 'start', 'end', 'points' keys
        seed: Random seed for reproducibility (initial value perturbations)
        param_seed: Random seed for parameter perturbations (uses seed if not provided)
        solver_type: Type of solver ('scipy' or 'roadrunner')
        jit: Whether to use JIT compilation (only for scipy solver)
        verbose: Whether to show progress bar
        require_all_successful: Whether to require all samples to succeed (default: False)
        return_details: If True, returns extended data structure with intermediate datasets (default: True)
        capture_all_species: If True, captures timecourses for all species in returned data.
        as_pandas: If True (default), returns pandas DataFrame for X and Series for y with feature names.
                  If False, returns numpy arrays for X and y.
        return_timecourse: If True, returns dictionary with X, y, timecourse, parameters, and metadata (default: False)

    Returns:
        If return_timecourse=False:
            If as_pandas=True: Tuple of (X, y) where X is pandas DataFrame with feature names
                and y is pandas Series
            If as_pandas=False: Tuple of (X, y) where X is numpy array (n_samples, n_features)
                and y is numpy array (n_samples,)
        If return_timecourse=True: Dictionary with keys:
            - 'X': Feature matrix (basal state values)
            - 'y': Target values
            - 'timecourse': Timecourse simulation data (DataFrame with numpy arrays)
            - 'parameters': Kinetic parameters dataframe (None if not perturbed)
            - 'metadata': Dictionary with metadata about generation process

    Raises:
        ValueError: If cell_model is not compiled or has invalid parameters
    """
    from .utils.data_generation_helpers import make_data

    if not cell_model._compiled:
        raise ValueError("cell_model must be compiled. Call cell_model.compile() first.")

    # Set default simulation parameters (use cell_model's simulation_end)
    if simulation_params is None:
        simulation_params = {'start': 0, 'end': cell_model._simulation_end, 'points': 101}
    
    # Validate simulation parameters
    if 'start' not in simulation_params or 'end' not in simulation_params or 'points' not in simulation_params:
        raise ValueError('simulation_params must contain "start", "end", and "points" keys')

    # Set default perturbation parameters for conserve_rules
    if perturbation_params is None:
        perturbation_params = {'shape': 0.5, 'base_shape': 0.01, 'max_shape': 0.5}
    
    # Auto-extract parameter values from cell model if not provided
    if parameter_values is None:
        parameter_values = cell_model.model.get_parameters()
    
    # Set default parameter perturbation parameters
    if param_perturbation_params is None:
        param_perturbation_params = {'shape': 0.1}
    
    # Pass model_spec to conserve_rules for auto-generation of species ranges
    # Note: This is optional - if model_spec is not provided, conserve_rules will use
    # the species_range dictionary if provided, or auto-generate from initial_values
    if perturbation_type == 'conserve_rules' and 'model_spec' not in perturbation_params and 'species_range' not in perturbation_params:
        perturbation_params = perturbation_params.copy()
        perturbation_params['model_spec'] = cell_model.spec

    # Get initial values (excluding drugs)
    initial_values = cell_model.get_initial_values(exclude_drugs=True)
    # Exclude any species with 'a' at the end of their name (outcome species and activated species)
    # Since they should not be perturbed initially
    initial_values = {k: v for k, v in initial_values.items() if not k.endswith('a')}
    initial_values.pop('O')

    # Create and compile solver
    if solver_type == 'scipy':
        solver = ScipySolver()
        antimony_str = cell_model.model.get_antimony_model()
        solver.compile(antimony_str, jit=jit)
    elif solver_type == 'roadrunner':
        solver = RoadrunnerSolver()
        sbml_str = cell_model.model.get_sbml_model()
        solver.compile(sbml_str)
    else:
        raise ValueError(f"Unsupported solver_type: {solver_type}. Use 'scipy' or 'roadrunner'")

    # Generate feature and target data using make_data
    result = make_data(
        initial_values=initial_values,
        perturbation_type=perturbation_type,
        perturbation_params=perturbation_params,
        n_samples=n,
        solver=solver,
        parameter_values=parameter_values,
        param_perturbation_type=param_perturbation_type,
        param_perturbation_params=param_perturbation_params,
        simulation_params=simulation_params,
        target_method="fold_change_drug",
        seed=seed,
        param_seed=param_seed,
        require_all_successful=require_all_successful,
        return_details=return_details,
        capture_all_species=capture_all_species,
        outcome_var=target_specie,
        verbose=verbose,
        n_cores=n_cores,
    )

    if return_details:
        # Return extended data structure
        # Use 'features' (perturbed initial values) as X, not 'basal_data' (pre-drug simulation state)
        X = result['features'].copy()
        if exclude_activated_from_features:
            X = X[[col for col in X.columns if not col.endswith('a')]]
        if exclude_outcome_from_features:
            X = X.drop(columns=['O'], errors='ignore')
        y = pd.Series(result['targets'].iloc[:, 0].values, name=target_specie, index=result['targets'].index, dtype=float)
        if not as_pandas:
            X = X.values.astype(np.float64)
            y = y.values.ravel().astype(np.float64)

        if return_timecourse:
            return {
                'X': X,
                'y': y,
                'timecourse': result['timecourse'],
                'parameters': result['parameters'],
                'metadata': result['metadata']
            }
        return X, y
    else:
        raise NotImplementedError("return_details=False is not implemented in this version.")


__all__ = ['VirtualCell', 'Builder', 'make_dataset_drug_response']



from typing import List, Dict, Optional, Tuple
import numpy as np
import logging
from .BaseSpec import BaseSpec
from .Regulation import Regulation
from .Drug import Drug
from .MichaelisNetworkSpec import MichaelisNetworkSpec
from ..ModelBuilder import ModelBuilder


class DegreeInteractionSpec(MichaelisNetworkSpec):
    """
    Multi-degree drug interaction network specification.
    
    Generates networks with hierarchical degrees of interaction:
    - Degree 1: D -> {R} -> {I} -> O (central pathway)
    - Degrees 2+: R -> I only (independent cascades influencing lower degrees)
    
    Each degree contains multiple cascades defined by degree_cascades list.
    Higher degrees provide feedback regulation to immediate lower degrees.
    
    Example:
        spec = DegreeInteractionSpec(degree_cascades=[1, 2, 5, 10])
        spec.generate_specifications(random_seed=42)
        model = spec.generate_network("multi_degree_network")
    """
    
    def __init__(self, degree_cascades: List[int], critical_pathways: Optional[int] = None):
        """
        Initialize degree interaction specification.
        
        Args:
            degree_cascades: List of cascade counts per degree, e.g., [1, 2, 5, 10]
            critical_pathways: Optional override for degree 1 cascade count
        """
        super().__init__()
        # Validate inputs before accessing degree_cascades[0]
        if len(degree_cascades) < 1:
            raise ValueError("degree_cascades must contain at least one element")
        
        self.degree_cascades = degree_cascades
        self.critical_pathways = critical_pathways or degree_cascades[0]
        self.degree_species: Dict[int, Dict[str, List[str]]] = {}  # {degree: {'R': [...], 'I': [...]}}
        self.degree_regulations: Dict[int, List[Regulation]] = {}  # Regulations per degree
        self._logger = logging.getLogger(__name__)
        
        if self.critical_pathways != degree_cascades[0]:
            self._logger.warning(f"critical_pathways ({self.critical_pathways}) differs from degree_cascades[0] ({degree_cascades[0]})")
    
    def generate_species_names(self) -> None:
        """
        Generate species names for all degrees based on cascade counts.
        Follows naming convention: R{degree}_{index}, I{degree}_{index}
        """
        self.degree_species.clear()
        
        for deg_idx, cascade_count in enumerate(self.degree_cascades):
            degree = deg_idx + 1  # 1-indexed degrees
            r_species = []
            i_species = []
            
            for cascade_idx in range(cascade_count):
                r_name = f"R{degree}_{cascade_idx + 1}"
                i_name = f"I{degree}_{cascade_idx + 1}"
                r_species.append(r_name)
                i_species.append(i_name)
                
                # Add to global species list
                self.add_species(r_name, group=f"degree_{degree}")
                self.add_species(i_name, group=f"degree_{degree}")
            
            self.degree_species[degree] = {'R': r_species, 'I': i_species}
        
        # Add outcome species O (only one)
        self.add_species("O", group="outcome")
    
    def generate_ordinary_regulations(self) -> None:
        """
        Generate ordinary regulations within each cascade:
        - Degree 1: R_i -> I_i, I_i -> O
        - Degrees 2+: R_i -> I_i only
        """
        self.degree_regulations.clear()
        
        for degree, species_dict in self.degree_species.items():
            degree_regs = []
            r_species = species_dict['R']
            i_species = species_dict['I']
            
            # Each cascade: R_i -> I_i
            for r_name, i_name in zip(r_species, i_species):
                reg = Regulation(from_specie=r_name, to_specie=i_name, reg_type='up')
                self.regulations.append(reg)
                degree_regs.append(reg)
            
            # Degree 1: I_i -> O
            if degree == 1:
                for i_name in i_species:
                    reg = Regulation(from_specie=i_name, to_specie='O', reg_type='up')
                    self.regulations.append(reg)
                    degree_regs.append(reg)
            
            self.degree_regulations[degree] = degree_regs
        
        # Track ordinary regulations (non-feedback)
        self.ordinary_regulations = self.regulations.copy()
    
    def generate_feedback_regulations(self, random_seed: Optional[int] = None, 
                                     feedback_density: float = 0.5) -> None:
        """
        Generate feedback regulations between adjacent degrees following structured rules.
        
        Rules:
        1. Downward regulations (n -> n-1): MANDATORY
           - For each cascade i in degree n > 1:
             - Regulator: I{n}_i (I species of that cascade)
             - Target: Random R or I species in degree (n-1)
             - Type: Random 'up' or 'down'
        
        2. Upward regulations (n-1 -> n): DENSITY-CONTROLLED
           - For each cascade i in degree n > 1:
             - Create candidate regulation:
               - Regulator: Random R or I species in degree (n-1)
               - Target: Random R{n}_i or I{n}_i
               - Type: Random 'up' or 'down'
           - Trim from outermost degrees first:
             - Keep only floor(feedback_density * cascade_count) upward regulations per degree
             - Randomly select which cascades keep their upward regulation
        
        Args:
            random_seed: Random seed for reproducibility
            feedback_density: Proportion of upward regulations to keep (0-1)
                - 0: No upward regulations
                - 1: All cascades get upward regulation
                - (0,1): Random subset based on density
        """
        if random_seed is not None:
            rng = np.random.default_rng(random_seed)
        else:
            rng = np.random.default_rng()
        
        # Clear existing feedback regulations
        self.feedback_regulations = []
        
        # Store candidate upward regulations by degree for trimming
        upward_candidates_by_degree: Dict[int, List[Tuple[int, Regulation]]] = {}
        # (degree, (cascade_index, regulation))
        
        # Helper function to check if regulation already exists
        def regulation_exists(regulator: str, target: str) -> bool:
            return any(
                r.from_specie == regulator and r.to_specie == target
                for r in self.regulations + self.feedback_regulations
            )
        
        # Generate regulations for each degree n > 1
        for degree in range(2, len(self.degree_cascades) + 1):
            prev_degree = degree - 1
            
            # Get species from both degrees
            prev_r_species = self.degree_species[prev_degree]['R']
            prev_i_species = self.degree_species[prev_degree]['I']
            prev_all_species = prev_r_species + prev_i_species
            
            curr_r_species = self.degree_species[degree]['R']
            curr_i_species = self.degree_species[degree]['I']
            curr_all_species = curr_r_species + curr_i_species
            
            if not prev_all_species or not curr_all_species:
                continue
            
            cascade_count = len(curr_r_species)  # Should be same as len(curr_i_species)
            upward_candidates_by_degree[degree] = []
            
            # 1. MANDATORY DOWNWARD REGULATIONS (n -> n-1)
            for i in range(cascade_count):
                regulator = curr_i_species[i]  # I{degree}_{i+1}
                
                # Randomly select target from previous degree
                target = rng.choice(prev_all_species)
                reg_type = rng.choice(['up', 'down'])
                
                # Ensure no duplicate regulation
                attempts = 0
                while regulation_exists(regulator, target) and attempts < 10:
                    target = rng.choice(prev_all_species)
                    attempts += 1
                
                reg = Regulation(from_specie=regulator, to_specie=target, reg_type=reg_type)
                self.regulations.append(reg)
                self.feedback_regulations.append(reg)
            
            # 2. CREATE CANDIDATE UPWARD REGULATIONS (n-1 -> n)
            for i in range(cascade_count):
                # Randomly select regulator from previous degree
                regulator = rng.choice(prev_all_species)
                
                # Randomly select target from current cascade (R or I)
                target_options = [curr_r_species[i], curr_i_species[i]]
                target = rng.choice(target_options)
                
                reg_type = rng.choice(['up', 'down'])
                
                # Ensure no duplicate regulation
                attempts = 0
                while regulation_exists(regulator, target) and attempts < 10:
                    regulator = rng.choice(prev_all_species)
                    target = rng.choice(target_options)
                    attempts += 1
                
                reg = Regulation(from_specie=regulator, to_specie=target, reg_type=reg_type)
                upward_candidates_by_degree[degree].append((i, reg))
        
        # 3. TRIM UPWARD REGULATIONS FROM OUTERMOST DEGREES FIRST
        # Process degrees from highest to lowest
        degrees_sorted = sorted(upward_candidates_by_degree.keys(), reverse=True)
        
        for degree in degrees_sorted:
            candidates = upward_candidates_by_degree[degree]
            if not candidates:
                continue
            
            cascade_count = len(self.degree_species[degree]['R'])
            
            # Calculate how many upward regulations to keep
            if feedback_density <= 0:
                # No upward regulations
                continue
            elif feedback_density >= 1:
                # Keep all upward regulations
                num_to_keep = len(candidates)
            else:
                # Keep proportion based on density
                num_to_keep = max(1, int(feedback_density * cascade_count))
                num_to_keep = min(num_to_keep, len(candidates))
            
            # Randomly select which cascades keep their upward regulation
            # We want to select num_to_keep unique cascade indices
            cascade_indices = list(set(idx for idx, _ in candidates))
            if len(cascade_indices) > num_to_keep:
                keep_indices = set(rng.choice(cascade_indices, size=num_to_keep, replace=False))
            else:
                keep_indices = set(cascade_indices)
            
            # Add kept regulations
            for cascade_idx, reg in candidates:
                if cascade_idx in keep_indices:
                    self.regulations.append(reg)
                    self.feedback_regulations.append(reg)
            
            self._logger.debug(
                f"Degree {degree}: Kept {len(keep_indices)}/{cascade_count} upward regulations "
                f"(density={feedback_density})"
            )
        
        self._logger.info(
            f"Generated {len(self.feedback_regulations)} feedback regulations "
            f"(density={feedback_density})"
        )
    
    def generate_specifications(self, random_seed: Optional[int] = None, 
                               feedback_density: float = 0.5, **kwargs):
        """
        Generate complete network specifications.
        
        Args:
            random_seed: Random seed for reproducibility
            feedback_density: Proportion of feedback connections to create (0-1)
            **kwargs: Additional arguments (unused, for compatibility)
        """
        # Clear existing state
        self.regulations = []
        self.species_list = []
        self.species_groups = {}
        self.ordinary_regulations = []
        self.feedback_regulations = []
        self.drugs = []
        self.drug_values = {}
        
        # Generate species and regulations
        self.generate_species_names()
        self.generate_ordinary_regulations()
        self.generate_feedback_regulations(random_seed, feedback_density)
        
        self._logger.info(
            f"Generated {len(self.degree_cascades)}-degree network with "
            f"{len(self.species_list)} species and {len(self.regulations)} regulations"
        )
    
    def add_drug(self, drug: Drug, value: Optional[float] = None):
        """
        Add a drug to the model, validating it only targets degree 1 R species.
        
        Args:
            drug: Drug object containing name, start_time, default_value, and regulations
            value: Optional override for drug value
            
        Raises:
            ValueError: If drug targets non-degree-1 species or non-R species
        """
        # Validate drug targets
        for target_species in drug.regulation:
            # Check if target is a degree 1 R species
            is_degree1_r = any(
                target_species == r_name 
                for r_name in self.degree_species.get(1, {}).get('R', [])
            )
            
            if not is_degree1_r:
                raise ValueError(
                    f"Drug {drug.name} can only target degree 1 R species. "
                    f"Invalid target: {target_species}"
                )
        
        # Call parent method
        super().add_drug(drug, value)
    
    def get_species_by_degree(self, degree: int, species_type: str = 'all') -> List[str]:
        """
        Get species for a specific degree.
        
        Args:
            degree: Degree number (1-indexed)
            species_type: 'R', 'I', or 'all'
            
        Returns:
            List of species names
        """
        if degree not in self.degree_species:
            return []
        
        species_dict = self.degree_species[degree]
        if species_type == 'R':
            return species_dict['R']
        elif species_type == 'I':
            return species_dict['I']
        elif species_type == 'all':
            return species_dict['R'] + species_dict['I']
        else:
            raise ValueError("species_type must be 'R', 'I', or 'all'")
    
    def get_regulations_by_degree(self, degree: Optional[int] = None) -> List[Regulation]:
        """
        Get regulations for a specific degree or all degrees.
        
        Args:
            degree: Degree number (1-indexed) or None for all
            
        Returns:
            List of Regulation objects
        """
        if degree is None:
            return self.regulations.copy()
        elif degree in self.degree_regulations:
            return self.degree_regulations[degree].copy()
        else:
            return []
    
    def get_total_species_count(self) -> int:
        """Get total number of species in the network."""
        return len(self.species_list)
    
    def get_total_cascades(self) -> int:
        """Get total number of cascades across all degrees."""
        return sum(self.degree_cascades)
    
    def __str__(self) -> str:
        """String representation of the specification."""
        return (f"DegreeInteractionSpec(degrees={len(self.degree_cascades)}, "
                f"cascades={self.degree_cascades}, "
                f"species={self.get_total_species_count()}, "
                f"regulations={len(self.regulations)})")



from typing import Union, Tuple, List 

class ReactionArchtype:

    '''
    ReactionArchtype is a factory class that generates Reaction objects
        name: str, 
        reactants: Tuple[str], 
        products: Tuple[str], 
        parameters: Tuple[str], 
        rate_law: str, 
        extra_states: Tuple[str] = (),
        assume_parameters_values: Union[dict, tuple] = None,
        assume_reactant_values: Union[dict, tuple] = None,
        assume_product_values: Union[dict, tuple] = None
    '''

    def __init__(self, 
        name: str, 
        reactants: Tuple[str], 
        products: Tuple[str], 
        parameters: Tuple[str], 
        rate_law: str, 
        extra_states: Tuple[str] = (),
        assume_parameters_values: Union[dict, tuple, None] = None,
        assume_reactant_values: Union[dict, tuple, None] = None,
        assume_product_values: Union[dict, tuple, None] = None,
        reversible: bool = False,
        reverse_rate_law: str = None):

        assert len(reactants) == len(set(reactants)), 'reactants must be unique'
        assert len(products) == 0 or len(products) == len(set(products)), 'products must be unique'
        assert len(parameters) == len(set(parameters)), 'parameters must be unique'
        assert len(extra_states) == len(set(extra_states)), 'extra_states must be unique'
        assert reversible is False or reverse_rate_law is not None, 'reverse_rate_law must be provided if reversible is True'
        
        self.name = name
        self.reactants = reactants
        self.products = products
        self.parameters = parameters
        self.extra_states = extra_states
        self.rate_law = rate_law

        # reverse reaction between reactants and products
        self.reversible = reversible
        self.reverse_rate_law = reverse_rate_law

        self.reactants_count = len(reactants)
        self.products_count = len(products)
        self.state_variables_count = len(reactants) + len(products)
        self.extra_states_count = len(extra_states)
        self.parameters_count = len(parameters)

        if reversible:
            self.validate_rate_laws(rate_law, reverse_rate_law)
        else:
            self.validate_rate_law(rate_law)
        self.validate_parameters(assume_parameters_values, parameters)
        self.validate_reactant_values(assume_reactant_values, reactants)
        self.validate_product_values(assume_product_values, products)

        # assumed values are used to fill in the missing values in the rate law
        # always update None or tuple into dict objects, so that Reaction class
        # only uses dict to handle assumed values logic 
        self.assume_parameters_values = self._convert_tuple_to_dict(assume_parameters_values, parameters)
        self.assume_reactant_values = self._convert_tuple_to_dict(assume_reactant_values, reactants)
        self.assume_product_values = self._convert_tuple_to_dict(assume_product_values, products)



    def _convert_tuple_to_dict(self, tuple: Union[dict, tuple, None], reference_tuple: Tuple[str]) -> dict:
        '''
        this function converts a tuple to a dict by 
        referring i-th element in the tuple to the i-th element in the reference_tuple
        '''
        # do not do anything if tuple is dict 
        if isinstance(tuple, dict):
            return tuple

        if tuple is None:
            return {}
        elif len(tuple) == 0: 
            return {}

        new_dict = {}
        for idx, value in enumerate(tuple):
            key = reference_tuple[idx]
            new_dict[key] = value
        
        return new_dict


    def validate_rate_law(self, rate_law) -> bool:
        '''
        this function validates the rate_law against the archtype
            It expects the rate_law to contain the parameters
            It expects the rate_law to contain the extra states
            ^ copilot generated
        '''

        for parameter in self.parameters:
            if rate_law.find(parameter) == -1:
                raise ValueError(f'{parameter} is not in the rate law')

        for extra_state in self.extra_states:
            if rate_law.find(extra_state) == -1:
                raise ValueError(f'{extra_state} is not in the rate law')

        return True

    def validate_rate_laws(self, rate_law, reverse_rate_law) -> bool:
        '''
        this function validates the rate_law against the archtype
            It expects the rate_law to contain the parameters
            It expects the rate_law to contain the extra states
            ^ copilot generated
        '''

        for parameter in self.parameters:
            if rate_law.find(parameter) == -1 and reverse_rate_law.find(parameter) == -1:
                raise ValueError(f'{parameter} is not in the rate law or reverse rate law')

        for extra_state in self.extra_states:
            if rate_law.find(extra_state) == -1 and reverse_rate_law.find(extra_state) == -1:
                raise ValueError(f'{extra_state} is not in the rate law or reverse rate law')

        return True

    def _abstract_validate_values(self, values: Union[tuple, dict, None], contained_list: Tuple) -> bool:
        
        '''
        this function validates the values against the archtype
        if values is tuple: 
            then it needs to be smaller than the number of [reactants, products, parameters] in the archtype
            assume first element in tuple matches with first element in self.parameters

        if values is dict:
            then it needs to be smaller than the number of [reactants, products, parameters] in the archtype
            cannot have dict keys that are not in self.parameters 
        
        '''

        if values is None:
            return True

        if len(values) != len(contained_list):
                raise ValueError(f'length of {values} is not the same as the archtype {contained_list}')

        if isinstance(values, dict):
            for key in values:
                if key not in contained_list:
                    raise ValueError(f'{key} is not in the {contained_list}')
        
        return True

    def validate_parameters(self, parameter_values, parameters_list) -> bool:

        '''
        Delegate to abstract implementation at _abstract_validate_values 
        '''

        return self._abstract_validate_values(parameter_values, parameters_list)


    def validate_reactant_values(self, reactant_values, reactant_list) -> bool:

        '''
        Delegate to abstract implementation at _abstract_validate_values 
        '''

        return self._abstract_validate_values(reactant_values, reactant_list)


    def validate_product_values(self, product_values, products_list) -> bool:

        '''
        Delegate to abstract implementation at _abstract_validate_values 
        '''

        return self._abstract_validate_values(product_values, products_list)
            

    def __str__(self) -> str:
        return f'{self.name} {self.reactants} {self.products} {self.parameters} {self.rate_law}'


# imports 
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm

def last_time_point_method(time_course_data, selected_species=None):
    if selected_species is None:
        selected_species = time_course_data.columns
    else:
        selected_species = selected_species
    selected_time_course_data = time_course_data[selected_species]
    last_time_points = selected_time_course_data.map(lambda x: x[-1])
    return last_time_points

def get_dynamic_features(col_data: pd.Series, 
                         normalise: bool = True,
                         abs_change_tolerance: float = 0.01) -> list:
    
    # dynamic features
    auc = np.trapz(col_data)
    max_val = np.max(col_data)
    max_time = np.argmax(col_data)
    min_val = np.min(col_data)
    min_time = np.argmin(col_data)

    median_val = np.median(col_data)

    # calculation of total fold change (tfc)
    start_val = col_data.iloc[0]
    end_val = col_data.iloc[-1]

    tfc = 0 
    if start_val == 0:
        tfc = 1000
    else:
        if end_val - start_val >= 0:
            tfc = (end_val - start_val) / start_val
        elif end_val - start_val < 0:
            if end_val == 0:
                tfc = -1000
            else:
                tfc = -((start_val - end_val) / end_val)

    # calculation of time to stability (tsv)
    tsv = len(col_data)
    while tsv > 1:
        if abs(col_data.iloc[tsv-1] - col_data.iloc[tsv-2]) < abs_change_tolerance:
            tsv -= 1
        else:
            tsv_value = col_data.iloc[tsv-1]
            break
    if tsv == 1:
        tsv_value = col_data.iloc[0]

    max_sim_time = len(col_data)
    n_auc = auc / max_sim_time
    n_max_time = max_time / max_sim_time
    n_min_time = min_time / max_sim_time
    n_tsv = tsv / max_sim_time
    
    if not normalise:
        # reset the values to the original values
        n_auc = auc
        n_max_time = max_time
        n_min_time = min_time
        n_tsv = tsv 

    dynamic_features = [n_auc, median_val, tfc, n_max_time,
                        max_val, n_min_time, min_val, n_tsv, tsv_value, start_val]

    return dynamic_features


def dynamic_features_method(
    time_course_data, selected_features=None, n_cores=1, verbose=0
):
    if selected_features is None:
        selected_features = time_course_data.columns
    else:
        selected_features = selected_features

    # use parallel processing to speed up the calculation with tqdm
    def process_data(row_data):
        row_dynamic_features = []
        for feature in selected_features:
            col_data = row_data[feature]
            # convert to pd Series for easier manipulation
            col_data = pd.Series(col_data)
            dyn_feats = get_dynamic_features(col_data)
            row_dynamic_features.extend(dyn_feats)
        return row_dynamic_features

    if n_cores > 1 or n_cores == -1:
        all_dynamic_features = Parallel(n_jobs=n_cores)(
            delayed(process_data)(time_course_data.iloc[i])
            for i in tqdm(
                range(time_course_data.shape[0]),
                desc="Calculating dynamic features",
                disable=verbose == 0,
            )
        )
    else:
        all_dynamic_features = []
        # iterate each row in the time course data
        for i in tqdm(
            range(time_course_data.shape[0]),
            desc="Calculating dynamic features",
            disable=verbose == 0,
        ):
            row_dynamic_features = []
            row_data = time_course_data.iloc[i]
            for feature in selected_features:
                col_data = row_data[feature]
                # convert to pd Series for easier manipulation
                col_data = pd.Series(col_data)
                dyn_feats = get_dynamic_features(col_data)
                row_dynamic_features.extend(dyn_feats)
            all_dynamic_features.append(row_dynamic_features)

    dynamic_feature_label = [
        "auc",
        "median",
        "tfc",
        "tmax",
        "max",
        "tmin",
        "min",
        "ttsv",
        "tsv",
        "init",
    ]
    new_df = pd.DataFrame(
        all_dynamic_features,
        columns=[
            s + "_" + dynamic_feature
            for s in selected_features
            for dynamic_feature in dynamic_feature_label
        ],
        index=time_course_data.index,
    )

    return new_df




from dataclasses import dataclass


@dataclass
class LinkedParameters: 

    '''
    represents parameters which are not bound to a single reaction object, 
    but are instead shared between multiple reactions. 
    '''

    def __init__(self, name: str, value: float):
        self.name = name
        self.value = value
